#ifndef DOMAIN_H
#define DOMAIN_H

#include "suif_inc.h"
#include "srel.h"

#include <map>
#include <vector>

using std::map;
//using std::vector;

extern const char * k_static_root;
extern const char * k_access;
extern const char * k_iteration_domain;
extern const char * k_domain_symtab;

typedef std::vector<var_sym *> var_sym_v; // why is std:: needed ?

struct Param_Iterator : public Global_Var_Decl {
    Param_Iterator(Const_String name) : Global_Var_Decl(name), iterator(false) {}
    bool iterator;
};

struct symtab_ann {
    global_symtab *	symtab;
    map<var_sym *, Param_Iterator *> vars;
    void add_var(var_sym *var);
};

class guard_constructor {
    int			verbose;
    symtab_ann *	sa;
    F_And *		f;
    F_Exists *		e;

public:
    Relation *	    	r;
    guard_constructor(symtab_ann *sa, int verbose) : sa(sa), verbose(verbose) {
	r = new Relation(0);
	e = r->add_exists();
	f = e->add_and();
    }
    ~guard_constructor() {
	delete r;
    }

    boolean construct_guard(instruction *ins);
    boolean fill_q(Constraint_Handle& h, operand op, int mult);
};

struct context_info {
    symtab_ann *sa;
    SRelation  *ctx;
    tree_proc  *tp;

    context_info() : sa(NULL), ctx(NULL), tp(NULL) {}
};

class domain_constructor {
    SRelation * 	c;
    symtab_ann *	sa;
    int			level;
    int			plevel;
    var_sym_v		its;
    var_sym_v		out;
    instruction *	root;
    int			verbose;
    int			oldstyle;
    std::vector<int>	prefix;
    tree_proc *		tp;

    SRelation * reorder(SRelation * c);
    named_lin_ineq * reorder(named_lin_ineq * c, int params = -1,
						 int output = 0);
public:
    domain_constructor(tree_proc *tp, int verbose, int oldstyle) : 
	tp(tp), verbose(verbose), oldstyle(oldstyle) {}
    void scan(context_info *ctx);
    boolean scan(tree_node_list * tp);
    boolean scan(tree_node * tp);
    void construct();
    void construct(tree_node_list * tp);
    void construct(tree_node * tp);
    void recurse(tree_node * tp);
    boolean map(tree_instr * i) { return map(i->instr(), NULL, -1); }
    boolean map(instruction * i, instruction *p, int pos);
    void add_access(dep_instr_annote *di, int pos, boolean write);
    bool check_access(instruction * i, instruction * p, int pos);
    boolean check_op(operand& op, instruction *p, int pos, boolean write = false);
    void visit(tree_node *t);
    boolean construct_guard(tree_if *ti, tree_node_list *tnl);
    void set_prefix_and_iterator(int prefix_value, var_sym *iterator = NULL);
};

struct domain_ann {
    instruction	   	*insn;
    SRelation      	*domain;
    int			 level;
    int			 plevel;
    var_sym_v		 its;
    std::vector<int>	 prefix;
    const char *	 function;
    int			 n_args;

    domain_ann(instruction *i, SRelation *d, int l, int pl, var_sym_v it,
		std::vector<int> p, const char *f, int n): 
	insn(i), domain(d), level(l), plevel(pl), its(it), prefix(p),
	function(f), n_args(n) {}
};

struct access_ann {
    boolean		 write;
    named_lin_ineq      *nlq;
    var_sym		*array;
    int		 	 output;
    int			 arg_pos; /* argument position in function call or -1 */

    access_ann(boolean w, named_lin_ineq *n, var_sym *a, int o, int pos) :
	write(w), nlq(n), array(a), output(o), arg_pos(pos) {}
};

void construct_domains(tree_proc * tp, int verbose, int oldstyle, context_info *ctx);
void init(int &argc, char *argv[]);

#endif
