#include "pdg.h"
#include <algorithm>
#include <vector>
#include "version.h"

using namespace pdg;
using std::vector;

static at_init register_history_entry(history_entry::register_type);

void history_entry::register_type()
{
    static struct_description d = { create };
    YAML_PTR_FIELD(d, history_entry, tool, str);
    YAML_PTR_FIELD(d, history_entry, version, str);
    YAML_SEQ_FIELD(d, history_entry, arguments, str);
    structure::register_type("perl/history_entry", &typeid(history_entry), &d.d);
}

static at_init register_pdg(PDG::register_type);

void PDG::register_type()
{
    static struct_description pdg_d = { create };
    YAML_INT_FIELD(pdg_d, PDG, dimension);
    YAML_INT_FIELD(pdg_d, PDG, cacheline);
    YAML_INT_FIELD(pdg_d, PDG, root);
    YAML_PTR_FIELD(pdg_d, PDG, placement, str);
    YAML_SEQ_FIELD(pdg_d, PDG, nodes, node);
    YAML_SEQ_FIELD(pdg_d, PDG, arrays, array);
    YAML_SEQ_FIELD(pdg_d, PDG, dependences, dependence);
    YAML_SEQ_FIELD(pdg_d, PDG, params, parameter);
    YAML_PTR_FIELD(pdg_d, PDG, context, UnionSet);
    YAML_SEQ_FIELD(pdg_d, PDG, history, history_entry);
    YAML_PTR_FIELD(pdg_d, PDG, name, str);
    YAML_SEQ_FIELD(pdg_d, PDG, statement_dimensions, int);
    structure::register_type("perl/PDG", &typeid(PDG), &pdg_d.d);
}

void PDG::dump(emitter& e)
{
    yll_emitter_set_transfer(e.e, "perl/PDG");
    structure::dump(e);
}

void PDG::add_history_line(const char *tool, int argc, char * argv[])
{
    history_entry *he = new history_entry();
    he->tool = new str(string(tool));
    he->version = new str(string(pdg_version()));
    for (int i = 1; i < argc; ++i)
	he->arguments.push_back(new str(argv[i]));
    history.push_back(he);
}

static at_init register_array(array::register_type);

void array::register_type()
{
    static struct_description d = { create };
    YAML_INT_FIELD(d, array, id);
    YAML_PTR_FIELD(d, array, name, str);
    YAML_SEQ_FIELD(d, array, dims, int);
    YAML_PTR_FIELD(d, array, map, UnionMap);
    YAML_PTR_FIELD(d, array, element_type, str);
    structure::register_type("perl/array", &typeid(array), &d.d);
}

static at_init register_parameter(parameter::register_type);

void parameter::register_type()
{
    static struct_description d = { create };
    YAML_INT_FIELD(d, parameter, id);
    YAML_PTR_FIELD(d, parameter, name, str);
    YAML_PTR_FIELD(d, parameter, value, integer);
    structure::register_type("perl/parameter", &typeid(parameter), &d.d);
}

static at_init register_composed_map(pdg::ComposedMap::register_type);

void pdg::ComposedMap::register_type()
{
    static struct_description d = { create };
    YAML_SEQ_FIELD(d, pdg::ComposedMap, pos, UnionMap);
    YAML_SEQ_FIELD(d, pdg::ComposedMap, neg, UnionMap);
    YAML_SEQ_FIELD(d, pdg::ComposedMap, mask, UnionSet);
    structure::register_type("perl/ComposedMap", 
			     &typeid(pdg::ComposedMap), &d.d);
}

static at_init register_reuse_pair(pdg::reuse_pair::register_type);

void pdg::reuse_pair::register_type()
{
    static struct_description d = { create };
    YAML_PTR_FIELD(d, pdg::reuse_pair, from, access);
    YAML_PTR_FIELD(d, pdg::reuse_pair, map, UnionMap);
    YAML_SEQ_FIELD(d, pdg::reuse_pair, ads, ComposedMap);
    YAML_PTR_FIELD(d, pdg::reuse_pair, ads_file, str);
    YAML_INT_FIELD(d, pdg::reuse_pair, count);
    structure::register_type("perl/reuse_pair", 
			     &typeid(pdg::reuse_pair), &d.d);
}

static at_init register_access(pdg::access::register_type);

void pdg::access::register_type()
{
    static const char *at_names[access::last];
    at_names[write] = "write";
    at_names[read] = "read";
    static struct_description d = { create };
    YAML_PTR_FIELD(d, pdg::access, array, pdg::array);
    YAML_ENUM_FIELD(d, pdg::access, type, at_names);
    YAML_PTR_FIELD(d, pdg::access, map, UnionMap);
    YAML_SEQ_FIELD(d, pdg::access, backward, reuse_pair);
    YAML_INT_FIELD(d, pdg::access, nr);
    structure::register_type("perl/access", &typeid(pdg::access), &d.d);
}

static at_init register_function_call(function_call::register_type);

void function_call::register_type()
{
    static struct_description s_d = { create };
    YAML_SEQ_FIELD(s_d, function_call, arguments, access);
    YAML_PTR_FIELD(s_d, function_call, name, str);
    structure::register_type("perl/function_call", &typeid(function_call), &s_d.d);
}

static at_init register_stat(statement::register_type);

void statement::register_type()
{
    static struct_description s_d = { create };
    YAML_INT_FIELD(s_d, statement, operation);
    YAML_INT_FIELD(s_d, statement, line);
    YAML_SEQ_FIELD(s_d, statement, accesses, access);
    YAML_PTR_FIELD(s_d, statement, top_function, function_call);
    structure::register_type("perl/statement", &typeid(statement), &s_d.d);
}

static at_init register_node(node::register_type);

void node::register_type()
{
    static struct_description node_d = { create };
    YAML_INT_FIELD(node_d, node, nr);
    YAML_PTR_FIELD(node_d, node, source, UnionSet);
    YAML_PTR_FIELD(node_d, node, scattering, UnionMap);
    YAML_PTR_FIELD(node_d, node, statement, pdg::statement);
    YAML_SEQ_FIELD(node_d, node, prefix, int);
    YAML_SEQ_FIELD(node_d, node, sources, pdg::dependence);
    structure::register_type("perl/node", &typeid(node), &node_d.d);
}

static at_init register_dependence(dependence::register_type);

void dependence::register_type()
{
    static const char *dt_names[dependence::last];
    dt_names[flow] = "flow";
    dt_names[anti] = "anti";
    dt_names[pn] = "pn";
    dt_names[pn_union] = "pn_union";
    dt_names[pn_part] = "pn_part";
    dt_names[pn_wire] = "pn_wire";
    dt_names[pn_hold] = "pn_hold";
    dt_names[pn_shift] = "pn_shift";
    dt_names[reuse] = "reuse";
    dt_names[uninitialized] = "uninitialized";
    static struct_description dependence_d = { create };
    YAML_PTR_FIELD(dependence_d, dependence, relation, UnionMap);
    YAML_PTR_FIELD(dependence_d, dependence, from, pdg::node);
    YAML_PTR_FIELD(dependence_d, dependence, to, pdg::node);
    YAML_PTR_FIELD(dependence_d, dependence, from_access, pdg::access);
    YAML_PTR_FIELD(dependence_d, dependence, to_access, pdg::access);
    YAML_ENUM_FIELD(dependence_d, dependence, type, dt_names);
    YAML_PTR_FIELD(dependence_d, dependence, array, pdg::array);
    YAML_INT_FIELD(dependence_d, dependence, reordering);
    YAML_INT_FIELD(dependence_d, dependence, multiplicity);
    YAML_PTR_FIELD(dependence_d, dependence, size, enumerator);
    YAML_PTR_FIELD(dependence_d, dependence, container, dependence);
    YAML_PTR_FIELD(dependence_d, dependence, value_size, integer);
    structure::register_type("perl/dependence", &typeid(dependence), &dependence_d.d);
}

static at_init register_enumerator(pdg::enumerator::register_type);

void pdg::enumerator::register_type()
{
    static struct_description e_d = { create };
    structure::register_type("perl/PDG::enumerator", &typeid(enumerator), &e_d.d);
}

void pdg::enumerator::init(SyckParser *p, SyckNode *n)
{
    s = new string(n->data.str->ptr, n->data.str->len);
}

void pdg::enumerator::dump(emitter& e)
{
    yll_emitter_set_transfer(e.e, "perl/PDG::enumerator");
    if (s)
	yll_emitter_write_string(e.e, s->c_str());
    else if (value >= 0)
	yll_emitter_write_int(e.e, value);
    else
	assert(0);
}

pdg::enumerator::~enumerator()
{
    if (s)
	delete s;
}

static at_init register_matrix(pdg::Matrix::register_type);

void pdg::Matrix::register_type()
{
    static struct_description m_d = { create };
    structure::register_type("perl/PDL::Matrix", &typeid(Matrix), &m_d.d);
}

void pdg::Matrix::init(SyckParser *p, SyckNode *n)
{
    el = vector<vector<int> >(n->data.list->idx);
    for (int i = 0; i < n->data.list->idx; i++) {
	serialize *val;
	SYMID oid = syck_seq_read(n, i);
	syck_lookup_sym( p, oid, (char **)&val );
	seq<serialize> * row = dynamic_cast<seq<serialize>*>(val);
	el[i] = vector<int>(row->size());
	for (int j = 0; j < row->size() ; ++j) {
	    str * s = dynamic_cast<str *>((*row)[j]);
	    el[i][j] = atoi(s->s.c_str());
	    delete s;
	}
	delete val;
    }
}

void pdg::Matrix::dump(emitter& e)
{
    yll_emitter_set_transfer(e.e, "perl/PDL::Matrix");
    yll_emitter_start_seq(e.e);
    for (int i = 0; i < el.size(); ++i) {
	yll_emitter_set_options(e.e, YLL_INLINE_SEQ);
	yll_emitter_start_seq(e.e);
	for (int j = 0; j < el[i].size(); ++j) {
	    yll_emitter_write_int(e.e, el[i][j]);
	}
	yll_emitter_end_seq(e.e);
    }
    yll_emitter_end_seq(e.e);
}

void pdg::Matrix::dump()
{
    for (int i = 0; i < el.size(); ++i) {
	for (int j = 0; j < el[i].size(); ++j) {
	    printf("%d ", el[i][j]);
	}
	puts("");
    }
}

pdg::Matrix::Matrix(Polyhedron *P)
{
    for (int i = 0; i < P->NbConstraints; ++i) {
	std::vector<int> v;
	for (int j = 0; j < 2+P->Dimension; ++j)
	    v.push_back(VALUE_TO_INT(P->Constraint[i][j]));
	el.push_back(v);
    }
}

pdg::Matrix::Matrix(::Matrix *M)
{
    for (int i = 0; i < M->NbRows; ++i) {
	std::vector<int> v;
	for (int j = 0; j < M->NbColumns; ++j)
	    v.push_back(VALUE_TO_INT(M->p[i][j]));
	el.push_back(v);
    }
}

pdg::Matrix::operator ::Matrix *() const
{
    assert(el.size() > 0);
    ::Matrix * m = Matrix_Alloc(el.size(), el[0].size());
    Value * p = m->p_Init;
    for (int i = 0; i < el.size(); ++i) {
	for (int j = 0; j < el[i].size(); ++j) {
	    value_set_si(*p++, el[i][j]);
	}
    }
    return m;
}

//static at_init register_hmatrix(HMatrix::register_type);

void HMatrix::register_type()
{
    static struct_description m_d = { create };
    structure::register_type("perl/PDL::HMatrix", &typeid(HMatrix), &m_d.d);
}

static at_init register_set(UnionSet::register_type);

void UnionSet::register_type()
{
    static struct_description s_d = { create };
    YAML_INT_FIELD(s_d, UnionSet, dim);
    YAML_SEQ_FIELD(s_d, UnionSet, constraints, Matrix);
    YAML_SEQ_FIELD(s_d, UnionSet, params, parameter);
    structure::register_type("perl/PDG::UnionSet", &typeid(UnionSet), &s_d.d);
}

void UnionSet::fill_fields(anchor_map *am)
{
    struct_description * sd = (struct_description *)(t_desc())[&typeid(*this)];
    map<const type_info *, description*>& desc = serialize::t_desc();

    map<string,serialize_p>::iterator i;
    for (i = d->begin(); i != d->end(); ) {
	map<string,serialize_p>::iterator j = i++;
	if (sd->m.find(j->first) != sd->m.end()) {
	    serialize *el = j->second;

	    if (sd->m[j->first]->set(am, this, j->second))
		d->erase(j);
	    /* Perhaps it was a Set instead of a UnionSet */
	    else if (j->first == string("constraints")) {
		Matrix *m = dynamic_cast<Matrix *>(j->second);
		if (m) {
		    d->erase(j);
		    constraints.push_back(m);
		}
	    }
	}
    }
}

void UnionSet::dump(emitter& e)
{
    yll_emitter_set_transfer(e.e, "perl/PDG::UnionSet");
    structure::dump(e);
}

UnionSet::UnionSet(int d, Polyhedron *P, seq<parameter>& p) : dim(d), params(p)
{
    for ( ; P; P = P->next) {
	pdg::Matrix *M = new pdg::Matrix(P);
	constraints.push_back(M);
    }
}

Polyhedron *pdg::constraints2polyhedron(seq<pdg::Matrix> const & c)
{
    Polyhedron * domain = NULL;
    Polyhedron **next = &domain;
    for (int i = 0; i < c.size(); ++i) {
	::Matrix * m(*c[i]);
	Polyhedron *P = Constraints2Polyhedron(m, POL_NO_DUAL);
	Matrix_Free(m);
	*next = P;
	next = &P->next;
    }
    return domain;
}

UnionSet::operator ::Polyhedron *() const
{
    if (constraints.size() == 1 && constraints[0]->el.size() == 0)
	return Universe_Polyhedron(dim + params.size());
    return constraints2polyhedron(constraints);
}

static at_init register_map(UnionMap::register_type);

void UnionMap::register_type()
{
    static struct_description m_d = { create };
    YAML_INT_FIELD(m_d, UnionMap, input);
    YAML_INT_FIELD(m_d, UnionMap, output);
    YAML_SEQ_FIELD(m_d, UnionMap, constraints, Matrix);
    YAML_SEQ_FIELD(m_d, UnionMap, params, parameter);
    structure::register_type("perl/PDG::UnionMap", &typeid(UnionMap), &m_d.d);
}

void UnionMap::fill_fields(anchor_map *am)
{
    struct_description * sd = (struct_description *)(t_desc())[&typeid(*this)];
    map<const type_info *, description*>& desc = serialize::t_desc();

    map<string,serialize_p>::iterator i;
    for (i = d->begin(); i != d->end(); ) {
	map<string,serialize_p>::iterator j = i++;
	if (sd->m.find(j->first) != sd->m.end()) {
	    serialize *el = j->second;

	    if (sd->m[j->first]->set(am, this, j->second))
		d->erase(j);
	    /* Perhaps it was a Map instead of a UnionMap */
	    else if (j->first == string("constraints")) {
		Matrix *m = dynamic_cast<Matrix *>(j->second);
		if (m) {
		    d->erase(j);
		    constraints.push_back(m);
		}
	    }
	}
    }
}

void UnionMap::dump(emitter& e)
{
    yll_emitter_set_transfer(e.e, "perl/PDG::UnionMap");
    structure::dump(e);
}

UnionMap::UnionMap(int in, int out, Polyhedron *P, seq<parameter>& p) : 
	input(in), output(out), params(p)
{
    for ( ; P; P = P->next) {
	pdg::Matrix *M = new pdg::Matrix(P);
	constraints.push_back(M);
    }
}

UnionMap::operator ::Polyhedron *() const
{
    return constraints2polyhedron(constraints);
}

pdg::UnionMap *pdg::scatter_dependence(pdg::dependence *dep)
{
    pdg::UnionMap *sw = dep->from ? dep->from->scattering : NULL;
    pdg::UnionMap *sr = dep->to->scattering;
    int sw_input = sw ? sw->input : 0;
    int sw_output = sw ? sw->output : 0;
    pdg::UnionMap *scattered = new pdg::UnionMap(sw_input, sr->input);
    int nparam = dep->relation->params.size();
    if (sw) {
	assert(sw->input == sr->input);
	assert(sw->constraints.size() == 1);
	assert(sw->params.size() == nparam);
    }
    assert(sr->constraints.size() == 1);
    pdg::Matrix *WM = sw ? sw->constraints[0] : NULL;
    pdg::Matrix *RM = sr->constraints[0];
    assert(sr->params.size() == nparam);
    for (int i = 0; i < dep->relation->constraints.size(); ++i) {
	pdg::Matrix *DM = dep->relation->constraints[i];
	pdg::Matrix *M = new pdg::Matrix;
	M->el = DM->el;
	for (int j = 0; j < M->el.size(); ++j)
	    M->el[j].insert(M->el[j].begin()+1, sw_input+sr->input, 0);
	assert(M->el.size());
	int cols = M->el[0].size();
	M->el.insert(M->el.begin(), sw_input+sr->input, vector<int>(cols));

	if (WM) {
	    int wcols = WM->el[0].size();
	    for (int j = 0; j < sw->input; ++j) {
		assert(WM->el[j][1+j] == 1);
		M->el[j][1+j] = 1;
		for (int k = 0; k < sw->output; ++k)
		    M->el[j][1+sw->input+sr->input+k] = WM->el[j][1+sw->input+k];
		for (int k = 0; k < nparam+1; ++k)
		    M->el[j][cols-(nparam+1)+k] = WM->el[j][wcols-(nparam+1)+k];
	    }
	}

	int rcols = RM->el[0].size();
	for (int j = 0; j < sr->input; ++j) {
	    assert(RM->el[j][1+j] == 1);
	    M->el[sw_input+j][1+sw_input+j] = 1;
	    for (int k = 0; k < sr->output; ++k)
		M->el[sw_input+j][1+sw_input+sr->input+sw_output+k] = 
			RM->el[j][1+sr->input+k];
	    for (int k = 0; k < nparam+1; ++k)
		M->el[sw_input+j][cols-(nparam+1)+k] = 
			RM->el[j][rcols-(nparam+1)+k];
	}

	scattered->constraints.push_back(M);
    }
    return scattered;
}
