#include <assert.h>
#include <stdlib.h>
#include "testlib.h"

using pdg::PDG;

static char *tmp_dir = NULL;

void remove_tmp_dir(void)
{
    char command[2*PATH_MAX+100];
    int n, ret;

    n = snprintf(command, sizeof(command), "rm -r %s", tmp_dir);
    assert(n < sizeof(command));
    ret = system(command);
    assert(WEXITSTATUS(ret) == 0);

    free(tmp_dir);
}

void construct_pdg(const char *base, const char *options)
{
    char command[2*PATH_MAX+100];
    char *srcdir;
    int n, ret;

    srcdir = getenv("srcdir");
    assert(srcdir);

    if (!tmp_dir) {
	tmp_dir = mkdtemp(strdup("isa_test_XXXXXX"));
	assert(tmp_dir);
	atexit(remove_tmp_dir);
    }

    n = snprintf(command, sizeof(command), "cp %s/inputs/%s.c %s", srcdir, base,
		 tmp_dir);
    assert(n < sizeof(command));
    ret = system(command);
    assert(WEXITSTATUS(ret) == 0);

    n = snprintf(command, sizeof(command), "../c2pdg %s %s/%s.c",
		 options ? options : "", tmp_dir, base);
    assert(n < sizeof(command));
    ret = system(command);
    assert(WEXITSTATUS(ret) == 0);
}

pdg::PDG *get_pdg(const char *base, const char *options)
{
    char command[2*PATH_MAX+100];
    int n;
    construct_pdg(base, options);

    FILE *in;
    n = snprintf(command, sizeof(command), "%s/%s.yaml", tmp_dir, base);
    assert(n < sizeof(command));
    in = fopen(command, "r");
    assert(in);

    PDG *pdg;
    pdg = yaml::Load<PDG>(in);
    assert(pdg);

    return pdg;
}

void construct_pn(const char *base, const char *options)
{
    char command[2*PATH_MAX+100];
    int n, ret;

    n = snprintf(command, sizeof(command), "../pn %s --input %s/%s.yaml",
					    options ? options : "",
					    tmp_dir, base);
    assert(n < sizeof(command));
    ret = system(command);
    assert(WEXITSTATUS(ret) == 0);
}

pdg::PDG *get_pn(const char *base, const char *options)
{
    char command[2*PATH_MAX+100];
    int n;

    construct_pn(base, options);
    FILE *in;
    n = snprintf(command, sizeof(command), "%s/%s_pn.yaml", tmp_dir, base);
    assert(n < sizeof(command));
    in = fopen(command, "r");
    assert(in);

    PDG *pdg;
    pdg = yaml::Load<PDG>(in);
    assert(pdg);

    return pdg;
}
