#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#include <syck.h>

#include "pdg.h"
#include "yaml.h"

#include <assert.h>

using namespace yaml;
using namespace pdg;

#define ARRAY_SIZE(a) (sizeof(a)/sizeof(*a))

class struct1 : public structure {
    static serialize *create() { return new struct1(); }
public:
    struct1 *loop;
    static void register_type();
    struct1() : loop(NULL) {}
};

at_init register_struct1(struct1::register_type);

void struct1::register_type()
{
    static struct_description s1_d = { create };
    YAML_PTR_FIELD(s1_d, struct1, loop, struct1);
    structure::register_type("struct1", &typeid(struct1), &s1_d.d);
}

class struct2 : public structure {
    static serialize *create() { return new struct2(); }
public:
    struct1 *test;
    static void register_type();
    struct2() : test(NULL) {}
};

at_init register_struct2(struct2::register_type);

void struct2::register_type()
{
    static struct_description s2_d = { create };
    YAML_PTR_FIELD(s2_d, struct2, test, struct1);
    structure::register_type("struct2", &typeid(struct2), &s2_d.d);
}


int main(int argc, char * argv[])
{
    assert (argc > 2);
//    int fd = open("../PDG/test2_t.yaml", O_RDONLY);
    FILE *fp = fopen(argv[1], "r");

    int type = atoi(argv[2]);

    switch (type) {
    case 0: {
	struct2 * pdg = yaml::Load<struct2>(fp);
	printf("root: %p\n", pdg);
	printf("test: %p\n", pdg->test);
	printf("test|loop: %p\n", pdg->test->loop);

	pdg->Dump();

	pdg->free();
	delete pdg;
	break;
    }
    case 1: {
	PDG * pdg = yaml::Load<PDG>(fp);

	printf("root: %p\n", pdg);
	printf("dim: %d\n", pdg->dimension);
	printf("placement: %s\n", pdg->placement->s.c_str());
	printf("nr nodes: %d\n", pdg->nodes.size());
	for (int i = 0; i < pdg->nodes.size(); ++i) {
	    printf("ptr: %p\n", pdg->nodes[i]);
	    printf("nr: %d\n", pdg->nodes[i]->nr);
	    printf("source: %p\n", pdg->nodes[i]->source);
	    printf("dim: %d\n", pdg->nodes[i]->source->dim);
	    printf("constraints: %p\n", pdg->nodes[i]->source->constraints[0]);
	    pdg->nodes[i]->source->constraints[0]->dump();
	    printf("scatt: %p\n", pdg->nodes[i]->scattering);
	    pdg->nodes[i]->scattering->constraints[0]->dump();
	    printf("stat: %p\n", pdg->nodes[i]->statement);
	    printf("operation: %d\n", pdg->nodes[i]->statement->operation);
	}

	pdg->Dump();

	pdg->free();
	delete pdg;
	break;
    }
    }

    fclose(fp);

    return 0;
}
